



pip check
IF %ERRORLEVEL% NEQ 0 exit /B 1
ckeygen --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
pyhtmlizer --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
tkconch --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
trial --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
twist --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
twistd --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
